/*************************************************************************
 *
 * hladinovy spinac
 * 
 * (c) Ales Povalac, alpov@alpov.net, http://www.alpov.net/
 * 
 *************************************************************************
 *
 * $Id: main.c 12 2008-07-25 12:48:03Z alpov $
 * 
 *************************************************************************/         


#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <util/atomic.h>

#define FALSE   0
#define TRUE    (!FALSE)

LOCKBITS = (LB_MODE_1 & BLB0_MODE_2 & BLB1_MODE_2);

FUSES = {
    .low = (FUSE_BODLEVEL & FUSE_BODEN & FUSE_SUT0 & FUSE_CKSEL3 & FUSE_CKSEL2 & FUSE_CKSEL1),
    .high = (FUSE_BOOTSZ0 & FUSE_BOOTSZ1 & FUSE_SPIEN),
};

#define set_relay(x)        { if (x) PORTD |= _BV(PD7); else PORTD &= ~_BV(PD7); }
#define get_ext_direct()    (PIND & _BV(PD5))
#define get_btn()           (!(PINC & _BV(PC2)))
#define set_led(x)          { if (x) PORTB |= _BV(PB0); else PORTB &= ~_BV(PB0); }

static unsigned int ee_thr_min EEMEM = 256;
static unsigned int ee_thr_max EEMEM = 512;
static unsigned int ee_tmax_ext EEMEM = 900;
static unsigned int ee_tmax_tank EEMEM = 60;
static unsigned char ee_active EEMEM = FALSE;

static unsigned int thr_min;
static unsigned int thr_max;
static unsigned int tmax_ext;
static unsigned int tmax_tank;

static volatile unsigned int timeout;


/*************************************************************************/
unsigned char get_ext(void)
{
    for (int i = 0; i < 10; i++) {
        if (get_ext_direct())
            return TRUE;
        _delay_ms(2);
    }
    return FALSE;
}


/*************************************************************************/
void morse_led(char c)
{
    #define T_DASH   300
    #define T_DOT    100
    #define T_SPACE  150

    wdt_reset();
    _delay_ms(T_SPACE);
    switch (c) {
        case 'H':
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        case 'M':
            set_led(TRUE); _delay_ms(T_DASH); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DASH); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        case 'T':
            set_led(TRUE); _delay_ms(T_DASH); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        case 'A':
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DASH); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        case 'N':
            set_led(TRUE); _delay_ms(T_DASH); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        case 'R':
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DASH); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        case 'S':
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            break;

        default:
            set_led(TRUE); _delay_ms(T_DOT); set_led(FALSE); _delay_ms(T_SPACE);
            break;
    }
    wdt_reset();
    _delay_ms(4*T_SPACE);
}


/*************************************************************************/
unsigned int get_level(void)
{
    unsigned int result = 0;
    
    for (int i = 0; i < 16; i++) {
        ADCSRA |= _BV(ADSC);
        loop_until_bit_is_clear(ADCSRA, ADSC);
        result += ADC;
        _delay_ms(1);
    }
    
    return (result / 16);
}


/*************************************************************************/
unsigned char get_active(void)
{
    unsigned char result;

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
        result = eeprom_read_byte(&ee_active);
    }
    return result;
}

/*************************************************************************/
void main_loop(void)
{
    wdt_reset();

    if (get_ext()) {
        if (get_level() > thr_max) {
            do {
                morse_led('M');
            } while (get_ext());
            return;
        }
        set_relay(TRUE);
        set_led(TRUE);
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
            timeout = tmax_ext;
        }
        do {
            wdt_reset();
            if (get_level() > thr_max) {
                set_relay(FALSE);
                do {
                    morse_led('M');
                } while (get_ext());
                return;
            }
            if (timeout == 0) {
                set_relay(FALSE);
                do {
                    morse_led('T');
                } while (get_ext());
                return;
            }
        } while (get_ext());
        set_relay(FALSE);
        set_led(FALSE);
        return;
    }
    else if (get_active()) {
        if (get_level() < thr_min) {
            set_relay(TRUE);
            set_led(TRUE);
            ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
                timeout = tmax_tank;
            }
            do {
                wdt_reset();
                if (timeout == 0) {
                    set_relay(FALSE);
                    do {
                        morse_led('T');
                    } while (get_active());
                    return;
                }
            } while ((get_level() < thr_max) && (get_active()));
            set_relay(FALSE);
            set_led(FALSE);
            return;
        }
        morse_led('A');
        return;
    }
    else {
        morse_led('N');
        return;
    }
}


/*************************************************************************/
void setup(void)
{
    morse_led('R');
    
    while (!get_btn())
        wdt_reset();

    eeprom_write_word(&ee_thr_min, get_level());
    morse_led('R');
    while (get_btn())
        wdt_reset();
    _delay_ms(50);
    set_relay(TRUE);
    set_led(TRUE);
    
    while (!get_btn())
        wdt_reset();

    eeprom_write_word(&ee_thr_max, get_level());
    set_relay(FALSE);
    set_led(FALSE);
    
    morse_led('R');
}


/*************************************************************************/
ISR(TIMER0_OVF_vect)
{
    static unsigned int cnt1 = 0;
    static unsigned int cnt2 = 0;

    if (get_btn()) {    // stisknuto
        cnt1++;
    } else {
        if (cnt1 > 2) {   // kratce
            eeprom_write_byte(&ee_active, !eeprom_read_byte(&ee_active));
        }
        cnt1 = 0;
    }
    
    cnt2++;
    if (cnt2 >= (F_CPU / 256 / 64)) {
        cnt2 = 0;
        if (timeout > 0)
            timeout--;
    }
}


/*************************************************************************/
int main(void)
{
    /* vstupne/vystupni porty */
    DDRB |= _BV(PB0);
    PORTC |= _BV(PC2);
    DDRD |= _BV(PD7);
    
    /* casovac pro tlacitko */
    TCCR0 = _BV(CS01) | _BV(CS00);
    TIMSK |= _BV(TOIE0);

    /* AD prevodnik */
    ADMUX = 6;
    ADCSRA = _BV(ADEN) | _BV(ADPS1) | _BV(ADPS0);
    
    /* watchdog */
    wdt_enable(WDTO_2S);
    
    if (get_btn()) {
        while (get_btn()) wdt_reset();
        setup();
        while (1) {}
    }

    /* dokonceni initu */
    thr_min = eeprom_read_word(&ee_thr_min);
    thr_max = eeprom_read_word(&ee_thr_max);
    tmax_ext = eeprom_read_word(&ee_tmax_ext);
    tmax_tank = eeprom_read_word(&ee_tmax_tank);
    sei();
    morse_led('H');

    while (1) {
        main_loop();
    }
}


